

/***************************************************************

   pdlsections.c

****************************************************************/

#include "pdl.h"      /* Data structure declarations */
#include "pdlcore.h"  /* Core declarations */



/*
   Code for subsection handling - extraction/insertion. Works
   for arbitrary dimensionality of data.
*/


/* Compute offset of (x,y,z,...) position in row-major list */

PDL_Indx pdl_get_offset(PDL_Indx* pos, PDL_Indx* dims, PDL_Indx *incs, PDL_Indx offset, PDL_Indx ndims) {
   PDL_Indx i;
   PDL_Indx result;
   result = offset;
   for (i=0; i<ndims; i++) {
       result = result + (pos[i]+((pos[i]<0)?dims[i]:0))*incs[i];
   }
   return result;
}

/* Check validity of section - return number of elements in it */

PDL_Indx pdl_validate_section( PDL_Indx* sec, PDL_Indx* dims, PDL_Indx ndims ){

   PDL_Indx i,start,end,count;

   count=1;

   for(i=0;i<ndims;i++){

       if (dims[i]<=0 || ndims==0)    /* Never happens :-) */
           croak("PDL object has a dimension <=0 !");

       start = sec[2*i];
       end   = sec[2*i+1];

       if (start<0 || end<0 || start>end || end>=dims[i])
            croak("Invalid subsection specified");

       count = count * (end-start+1);
   }
   return count;
}

/* Increment a position pointer array by one row */

void pdl_row_plusplus ( PDL_Indx* pos, PDL_Indx* dims, PDL_Indx ndims ) {

    PDL_Indx i;
    int noescape;

    i=1; noescape=1;

    while(noescape) {

       (pos[i])++;

       if (pos[i]==dims[i]) { /* Carry */
          if (i>=(ndims)-1)  {
             noescape = 0; /* Exit */
          }else{
             pos[i]=0;
             i++;
          }
       }else{
          noescape = 0;    /* Exit */
       }
    }
}

/* wrapper for pdl_at where only want first item, cf sclr_c */
PDL_Anyval pdl_at0( pdl* it ) {
    PDL_Indx nullp = 0;
    PDL_Indx dummyd = 1;
    PDL_Indx dummyi = 1;
    pdl_make_physvaffine( it );
    if (it->nvals < 1)
       croak("ndarray must have at least one element");
    return pdl_at(PDL_REPRP(it), it->datatype, &nullp, &dummyd,
            &dummyi, PDL_REPROFFS(it),1);
}

/* Return value at position (x,y,z...) */

PDL_Anyval pdl_at( void* x, int datatype, PDL_Indx* pos, PDL_Indx* dims, 
	PDL_Indx* incs, PDL_Indx offset, PDL_Indx ndims) {

    PDL_Indx i;
    PDL_Indx ioff;
    PDL_Anyval result = { -1, 0 };

    for(i=0; i<ndims; i++) { /* Check */

       /* leave pdl_get_offset to handle -ve offsets (i.e. from end of
          dimension), so elements of pos[] won't be changed */

       if(pos[i]<-dims[i] || pos[i]>=dims[i])
          croak("Position out of range");
    }

   ioff = pdl_get_offset(pos, dims, incs, offset, ndims);

   switch (datatype) {

   case PDL_B:
      {   
   
         PDL_Byte *xx = (PDL_Byte *) x;
         result.type = datatype;
         result.value.B = xx[ioff];
   
      }break;

   case PDL_S:
      {   
   
         PDL_Short *xx = (PDL_Short *) x;
         result.type = datatype;
         result.value.S = xx[ioff];
   
      }break;

   case PDL_US:
      {   
   
         PDL_Ushort *xx = (PDL_Ushort *) x;
         result.type = datatype;
         result.value.U = xx[ioff];
   
      }break;

   case PDL_L:
      {   
   
         PDL_Long *xx = (PDL_Long *) x;
         result.type = datatype;
         result.value.L = xx[ioff];
   
      }break;

   case PDL_IND:
      {   
   
         PDL_Indx *xx = (PDL_Indx *) x;
         result.type = datatype;
         result.value.N = xx[ioff];
   
      }break;

   case PDL_LL:
      {   
   
         PDL_LongLong *xx = (PDL_LongLong *) x;
         result.type = datatype;
         result.value.Q = xx[ioff];
   
      }break;

   case PDL_F:
      {   
   
         PDL_Float *xx = (PDL_Float *) x;
         result.type = datatype;
         result.value.F = xx[ioff];
   
      }break;

   case PDL_D:
      {   
   
         PDL_Double *xx = (PDL_Double *) x;
         result.type = datatype;
         result.value.D = xx[ioff];
   
      }break;

   case PDL_CF:
      {   
   
         PDL_CFloat *xx = (PDL_CFloat *) x;
         result.type = datatype;
         result.value.G = xx[ioff];
   
      }break;

   case PDL_CD:
      {   
   
         PDL_CDouble *xx = (PDL_CDouble *) x;
         result.type = datatype;
         result.value.C = xx[ioff];
   
      }break;

   default:
      croak ("Not a known data type code=%d",datatype);
   }

   return result;
}

/* Set value at position (x,y,z...) */

void pdl_set( void* x, int datatype, PDL_Indx* pos, PDL_Indx* dims, PDL_Indx* incs, PDL_Indx offs, PDL_Indx ndims, PDL_Anyval value){

    PDL_Indx i;
    PDL_Indx ioff;

    for(i=0; i<ndims; i++) { /* Check */

       if(pos[i]<-dims[i] || pos[i]>=dims[i])
          croak("Position out of range");
    }

   ioff = pdl_get_offset(pos, dims, incs, offs, ndims);

   switch (datatype) {

   case PDL_B:
      {   
   
         PDL_Byte *xx = (PDL_Byte *) x;
         ANYVAL_TO_CTYPE(xx[ioff], PDL_Byte, value);
   
      }break;

   case PDL_S:
      {   
   
         PDL_Short *xx = (PDL_Short *) x;
         ANYVAL_TO_CTYPE(xx[ioff], PDL_Short, value);
   
      }break;

   case PDL_US:
      {   
   
         PDL_Ushort *xx = (PDL_Ushort *) x;
         ANYVAL_TO_CTYPE(xx[ioff], PDL_Ushort, value);
   
      }break;

   case PDL_L:
      {   
   
         PDL_Long *xx = (PDL_Long *) x;
         ANYVAL_TO_CTYPE(xx[ioff], PDL_Long, value);
   
      }break;

   case PDL_IND:
      {   
   
         PDL_Indx *xx = (PDL_Indx *) x;
         ANYVAL_TO_CTYPE(xx[ioff], PDL_Indx, value);
   
      }break;

   case PDL_LL:
      {   
   
         PDL_LongLong *xx = (PDL_LongLong *) x;
         ANYVAL_TO_CTYPE(xx[ioff], PDL_LongLong, value);
   
      }break;

   case PDL_F:
      {   
   
         PDL_Float *xx = (PDL_Float *) x;
         ANYVAL_TO_CTYPE(xx[ioff], PDL_Float, value);
   
      }break;

   case PDL_D:
      {   
   
         PDL_Double *xx = (PDL_Double *) x;
         ANYVAL_TO_CTYPE(xx[ioff], PDL_Double, value);
   
      }break;

   case PDL_CF:
      {   
   
         PDL_CFloat *xx = (PDL_CFloat *) x;
         ANYVAL_TO_CTYPE(xx[ioff], PDL_CFloat, value);
   
      }break;

   case PDL_CD:
      {   
   
         PDL_CDouble *xx = (PDL_CDouble *) x;
         ANYVAL_TO_CTYPE(xx[ioff], PDL_CDouble, value);
   
      }break;

   default:
      croak ("Not a known data type code=%d",datatype);
   }
}



